"""removed ProcessUnits table

Revision ID: bcda1e0dd461
Revises: d53dfbe96f00
Create Date: 2024-11-27 11:22:00.975364

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'bcda1e0dd461'
down_revision: Union[str, None] = 'd53dfbe96f00'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('ProcessUnits')
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('ProcessUnits',
    sa.Column('engine_id', sa.VARCHAR(), nullable=False),
    sa.Column('state', sa.VARCHAR(), nullable=False),
    sa.Column('location', sa.VARCHAR(), nullable=False),
    sa.Column('runtime_msec', sa.INTEGER(), nullable=False),
    sa.Column('current_user_role', sa.VARCHAR(), nullable=False),
    sa.Column('id', sa.INTEGER(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###
