"""Added ErrorLog to RecentRun

Revision ID: d517729a9be0
Revises: 5dbc834cf70e
Create Date: 2024-02-12 14:22:09.812491

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'd517729a9be0'
down_revision: Union[str, None] = '5dbc834cf70e'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('RecentRunErrorLogs',
    sa.Column('run_id', sa.String(), nullable=False),
    sa.Column('error_log', sa.JSON(), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_RecentRunErrorLogs'))
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('RecentRunErrorLogs')
    # ### end Alembic commands ###
