"""change computer_name to engine_computer_name

Revision ID: 72449425de7c
Revises: 646d56d49b76
Create Date: 2024-01-30 10:42:08.070364

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '72449425de7c'
down_revision: Union[str, None] = '646d56d49b76'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('RecentRuns', 'computer_name', new_column_name='engine_computer_name')
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('RecentRuns', 'engine_computer_name', new_column_name='computer_name')
    # ### end Alembic commands ###
